#!/usr/bin/env python
#
# Nametable editor .sav to PNG converter using Python Imaging Library
# Copyright 2012 Damian Yerrick
#
# Copying and distribution of this file, with or without
# modification, are permitted in any medium without royalty
# provided the copyright notice and this notice are preserved.
# This file is offered as-is, without any warranty.
#
from __future__ import with_statement
from PIL import Image
import sys
from array import array
from binascii import a2b_hex

# Bisqwit's palette
nespalbytes = a2b_hex(
    '656565002d69131f7f3c137c600b62730a37710f075a1a00'
    '3428000b3400003c00003d10003840000000000000000000'
    'aeaeae0f63b34051d07841cca736a9c03470bd3c309f4a00'
    '6d5c00366d0007770400793d00727d000000000000000000'
    'fefeff5db3ff8fa1ffc890fff785faff83c0ff8b7fef9a49'
    'bdac2c85bc2f55c7533cc98c3ec2cd4e4e4e000000000000'
    'fefeffbcdfffd1d8ffe8d1fffbcdfdffcce5ffcfcaf8d5b4'
    'e4dca8cce3a9b9e8b8aee8d0afe5eab6b6b6000000000000'
)
nespal = [nespalbytes[i:i + 3]
          for i in range(0, len(nespalbytes), 3)]

def sliver_to_texels(lo, hi):
    return [((lo >> i) & 1) | (((hi >> i) & 1) << 1)
            for i in range(7, -1, -1)]

def tile_to_texels(chrdata):
    _stt = sliver_to_texels
    _o = ord
    return [_stt(_o(a), _o(b)) for (a, b) in zip(chrdata[0:8], chrdata[8:16])]

def add_attr_to_texels(texels, attr):
    attr = attr * 4
    return [[c and c + attr for c in row] for row in texels]

def chrbank_to_texels(chrdata):
    _ttt = tile_to_texels
    return [_ttt(chrdata[i:i + 16]) for i in range(0, len(chrdata), 16)]

def texels_to_pil(texels, tile_width=16):
    r8 = range(8)
    texels = [texels[i:i + tile_width]
              for i in range(0, len(texels), tile_width)]
    texels = ["".join(chr(c)
                      for tile in row for c in tile[y])
              for row in texels for y in range(8)]
    im = Image.fromstring('P', (8 * tile_width, len(texels)), ''.join(texels))
    im.putpalette('\x00\x00\x00\x66\x66\x66\xb2\xb2\xb2\xff\xff\xff'*64)
    return im

def decode_attribute_table(attrs):
    """Decode a 64-byte NES attribute table to a 16x15 grid."""

    from binascii import b2a_hex
    
    # linearize attributes
    attrs = [attrs[i:i + 8] for i in range(0, 64, 8)]
    attrs = [([((ord(c) >> 0) & 0x03, (ord(c) >> 2) & 0x03)
               for c in row],
              [((ord(c) >> 4) & 0x03, (ord(c) >> 6) & 0x03)
               for c in row])
             for row in attrs]
    attrs = [[subcol for byte in subrow for subcol in byte]
             for row in attrs for subrow in row]
    return attrs[:15]

def parse_argv(argv):
    from optparse import OptionParser
    parser = OptionParser(usage="usage: %prog [options] [-i] INFILE [-o] OUTFILE")
    parser.add_option("-i", "--image", dest="infilename",
                      help="read 16 KiB PPU dump from INFILE", metavar="INFILE")
    parser.add_option("-o", "--output", dest="outfilename",
                      help="write SAV data to OUTFILE", metavar="OUTFILE")
    parser.add_option("-P", "--pattern-table", "--chr", dest="pattable",
                      help="write tile sheet instead of screen, with colorset COLORSET (0-3)", metavar="COLORSET",
                      default=None, type="int")
    (options, args) = parser.parse_args(argv[1:])

    # Fill unfilled roles with positional arguments
    argsreader = iter(args)
    try:
        infilename = options.infilename
        if infilename is None:
            infilename = argsreader.next()
        outfilename = options.infilename
        if outfilename is None:
            outfilename = argsreader.next()
    except StopIteration:
        parser.error("not enough filenames")
    try:
        argsreader.next()
        parser.error("too many filenames")
    except StopIteration:
        pass

    if options.pattable not in (0, 1, 2, 3, None):
        parser.error("palette must be 0 to 3")
        
    return (infilename, outfilename, options.pattable)

def main(argv=None):
    (infilename, outfilename, write_pattable) = parse_argv(argv or sys.argv)
    print "infilename:", infilename
    print "outfilename:", outfilename
    print "write_pattable:", write_pattable
    with open(infilename, 'rb') as infp:
        sav = infp.read()
    chrdata = sav[0x0000:0x1000]
    palette = sav[0x1F00:0x1F20]
    rgbpalette = [nespal[ord(c) & 0x3F] for c in palette]
    tiles = chrbank_to_texels(chrdata)

    # Export CHR bank instead if desired
    if write_pattable is not None:
        tiles = texels_to_pil(tiles)
        subpal = rgbpalette[0:1] + rgbpalette[write_pattable*4+1:write_pattable*4+4]
        tiles.putpalette(''.join(subpal) * 64)
        tiles.save(outfilename)
        return
    
    nam = sav[0x1800:0x1C00]
    attrs = decode_attribute_table(nam[0x3C0:])
    r2 = (0,1)
    attrs = [[c for c in row for i in r2] for row in attrs for i in r2]
    nam = [[ord(c) for c in nam[i:i + 32]] for i in range(0, 960, 32)]
    tile = add_attr_to_texels(tiles[nam[2][3]], attrs[2][3])
    colortiles = [add_attr_to_texels(tiles[tn], cs)
                  for (nr, ar) in zip(nam, attrs) for (tn, cs) in zip(nr, ar)]
    im = texels_to_pil(colortiles, 32)
    im.putpalette(''.join(rgbpalette) * 8)
    im.save(outfilename)

if __name__=='__main__':
##    main(['sav2png.py', '../editor.sav', 'out.png', '--chr', '2'])
##    main(['sav2png.py', '../editor.sav', 'out.png'])
    main()
